from __future__ import print_function, division, absolute_import

import sys
import pickle
from copy import copy
from collections import namedtuple

import numpy as np
import pytest

from crick import SpaceSaving


data_f8 = np.random.RandomState(42).gamma(0.1, 0.1, size=10000).round(2) * 100
data_i8 = data_f8.astype('i8')
data_object = data_i8.astype(object)
data_string = np.array([str(i) for i in data_object], dtype=object)


def topk(data, k):
    unique, counts = np.unique(data, return_counts=True)
    inds = counts.argsort()[::-1]
    return unique[inds[:k]], counts[inds[:k]]


@pytest.mark.parametrize('dtype,data',
                         [(float, data_f8),
                          (int, data_i8),
                          (object, data_object),
                          (object, data_string),
                          (object, data_i8),
                          (float, data_i8)])
def test_space_saving(dtype, data):
    s = SpaceSaving(capacity=20, dtype=dtype)
    s.update(data)
    assert s.size() == s.capacity
    top = s.topk(10)
    vals, counts = topk(data, 10)
    assert (top['item'] == vals).all()
    assert (top['count'] == counts).all()


def test_add_matches_update():
    s = SpaceSaving(capacity=20, dtype=int)
    s2 = SpaceSaving(capacity=20, dtype=int)
    s.update(data_i8)
    for i in data_i8:
        s2.add(i)
    np.testing.assert_equal(s.counters(), s2.counters())


def test_init():
    s = SpaceSaving(capacity=10, dtype='f8')
    assert s.dtype == np.dtype('f8')
    assert s.capacity == 10
    assert SpaceSaving(dtype='i4').dtype == np.dtype('i8')

    for c in [-10, 0, np.nan]:
        with pytest.raises(ValueError):
            SpaceSaving(capacity=c)

    with pytest.raises(ValueError):
        SpaceSaving(dtype='S1')


def test_algorithm():
    s = SpaceSaving(capacity=5, dtype='i8')
    assert s.size() == 0
    assert len(s.counters()) == 0

    # Add 5 independent items
    for i in range(5):
        s.add(i)
    c = s.counters()
    assert len(c) == 5
    np.testing.assert_equal(c['item'], np.arange(5, dtype='i8'))
    np.testing.assert_equal(c['count'], 1)
    np.testing.assert_equal(c['error'], 0)

    # Add the tail one twice, should move to front
    s.add(4, 2)
    c = s.counters()
    np.testing.assert_equal(c['item'], np.array([4, 0, 1, 2, 3]))
    np.testing.assert_equal(c['count'], np.array([3, 1, 1, 1, 1]))
    np.testing.assert_equal(c['error'], 0)

    # Add a middle one, should move to 2nd
    s.add(2)
    c = s.counters()
    np.testing.assert_equal(c['item'], np.array([4, 2, 0, 1, 3]))
    np.testing.assert_equal(c['count'], np.array([3, 2, 1, 1, 1]))
    np.testing.assert_equal(c['error'], 0)

    # Add new element, tail should drop off
    s.add(5)
    c = s.counters()
    np.testing.assert_equal(c['item'], np.array([4, 2, 5, 0, 1]))
    np.testing.assert_equal(c['count'], np.array([3, 2, 2, 1, 1]))
    np.testing.assert_equal(c['error'], np.array([0, 0, 1, 0, 0]))

    # Update some more elements
    s.update([6, 7] * 5)
    c = s.counters()
    np.testing.assert_equal(c['item'], np.array([6, 7, 4, 2, 5]))
    np.testing.assert_equal(c['count'], np.array([6, 6, 3, 2, 2]))
    np.testing.assert_equal(c['error'], np.array([1, 1, 0, 0, 1]))


def test_object_reference_counting():
    s = SpaceSaving(capacity=5, dtype=object)

    data = ('this', 'should', 'have', 'no', 'other', 'refs')
    data2 = ('neither', 'should', 'this')
    data3 = ('added', 'later', 'to', 'check', 'swapping', 'counts')
    array_data2 = np.array([None], dtype=object)
    array_data2[0] = data2
    array_data3 = np.array([1, 2, 3, data3] * 2, dtype=object)

    # Adding increases the reference count
    orig = sys.getrefcount(data)
    s.add(data)
    after = sys.getrefcount(data)
    assert after == orig + 1

    # Updating increases the reference count
    orig2 = sys.getrefcount(data2)
    s.update(array_data2)
    after2 = sys.getrefcount(data2)
    assert after2 == orig2 + 1

    # Adding again doesn't change things
    s.add(data)
    assert sys.getrefcount(data) == after

    # Dropping decreases the reference count, swapped items
    # have reference counts increased
    orig3 = sys.getrefcount(data3)
    s.update(array_data3)
    after3 = sys.getrefcount(data3)
    assert after3 == orig3 + 1
    assert sys.getrefcount(data2) == orig2

    # check that deleting decreases the reference count
    del s
    assert sys.getrefcount(data) == orig
    assert sys.getrefcount(data2) == orig2
    assert sys.getrefcount(data3) == orig3


def test_object_hash_errors():
    s = SpaceSaving(capacity=5, dtype=object)
    data = ['lists', 'are', 'unhashable']
    orig = sys.getrefcount(data)

    with pytest.raises(TypeError):
        s.add(data)

    # nothing was added and no change in refcount
    assert s.size() == 0
    assert sys.getrefcount(data) == orig


def test_weights():
    s = SpaceSaving(capacity=5, dtype='i8')
    s.add(1, 10)
    c = s.counters()
    np.testing.assert_equal(c['item'], 1)
    np.testing.assert_equal(c['count'], 10)

    s.update([1, 2, 3], 5)
    c = s.counters()
    np.testing.assert_equal(c['item'], [1, 2, 3])
    np.testing.assert_equal(c['count'], [15, 5, 5])

    s.update(3, [5, 10, 5])
    c = s.counters()
    np.testing.assert_equal(c['item'], [3, 1, 2])
    np.testing.assert_equal(c['count'], [25, 15, 5])

    s.update([1, 2, 3], [1, 2, 3])
    c = s.counters()
    np.testing.assert_equal(c['item'], [3, 1, 2])
    np.testing.assert_equal(c['count'], [28, 16, 7])


def test_add_raises():
    s = SpaceSaving(capacity=5, dtype='i8')

    for x, c in [(1, 0), (1, -10), (1, np.nan), (np.nan, 1)]:
        with pytest.raises(ValueError):
            s.add(x, c)

    for x, c in [('foo', 1), (1, 'foo')]:
        with pytest.raises(TypeError):
            s.add(x, c)

    for x, c in [(1, np.inf), (1, -np.inf), (np.inf, 1), (-np.inf, 1)]:
        with pytest.raises(OverflowError):
            s.add(x, c)


def test_update_raises():
    s = SpaceSaving(capacity=5, dtype='i8')

    for x, c in [(1, 0), (1, -10)]:
        for x2, c2 in [([x], [c]), ([x], c), (x, [c])]:
            with pytest.raises(ValueError):
                s.update(x2, c2)

    for x, c in [(1, np.nan), (1, np.inf), (1, -np.inf), (1, 'foo'),
                 (np.nan, 1), (np.inf, 1), (-np.inf, 1), ('foo', 1)]:
        with pytest.raises(TypeError):
            for x2, c2 in [([x], [c]), ([x], c), (x, [c])]:
                s.update(x2, c2)


def test_topk_invariants():
    s = SpaceSaving(capacity=5, dtype='f8')
    s.update(data_f8)
    for k in [0, 5]:
        top = s.topk(k)
        assert isinstance(top, np.ndarray)
        dtype = np.dtype([('item', 'f8'), ('count', 'i8'), ('error', 'i8')])
        assert top.dtype == dtype
        assert len(top) == k
        assert (np.diff(top['count']) <= 0).all()

        top2 = s.topk(k, astuples=True)
        assert len(top2) == k
        np.testing.assert_equal(top['item'], [i.item for i in top2])
        np.testing.assert_equal(top['count'], [i.count for i in top2])
        np.testing.assert_equal(top['error'], [i.error for i in top2])

    with pytest.raises(ValueError):
        s.topk(-1)


@pytest.mark.parametrize('dtype', [float, int, object])
def test_serialize(dtype):
    empty = SpaceSaving(capacity=5, dtype=dtype)
    nonempty = SpaceSaving(capacity=5, dtype=dtype)
    # Results in nonuniform count and error, and non-sorted counters
    nonempty.update([1, 2, 3, 4, 5, 3, 3, 3, 6, 7])

    for s in [empty, nonempty]:
        s2 = pickle.loads(pickle.dumps(s))
        assert s2.capacity == s.capacity
        assert s2.size() == s.size()
        np.testing.assert_equal(s.counters(), s2.counters())
        # Add an element already in the table, to ensure hashing still works
        s.add(1)
        s2.add(1)
        np.testing.assert_equal(s.counters(), s2.counters())


Counter = namedtuple('Counter', ['item', 'count', 'error'])


def counter_sort_key(c):
    return (c.count, -c.error)


def merge(s1, s2):
    """A reimplementation of the merging algorithm, for testing purposes."""
    capacity = s1.capacity
    c1 = s1.counters()
    c2 = s2.counters()
    m1 = c1['count'][-1] if s1.size() == s1.capacity else 0
    m2 = c2['count'][-1] if s2.size() == s2.capacity else 0

    s1 = {i['item']: Counter(*i) for i in c1}
    s2 = {i['item']: Counter(*i) for i in c2}

    out = {}
    for item in set(s1).intersection(s2):
        out[item] = Counter(item,
                            s1[item].count + s2[item].count,
                            s1[item].error + s2[item].error)

    for item in set(s1).difference(s2):
        out[item] = Counter(item,
                            s1[item].count + m2,
                            s1[item].error + m2)

    for item in set(s2).difference(s1):
        out[item] = Counter(item,
                            s2[item].count + m1,
                            s2[item].error + m1)

    res = sorted(out.values(), key=counter_sort_key, reverse=True)
    return res[:capacity]


d1 = [1, 2, 3, 4, 5, 5, 5, 6]
d2 = [1, 2, 3, 5, 5, 6, 6, 7]

# s1 and s2 are designed to hit each case of the merge algorithm
s1 = SpaceSaving(capacity=5, dtype='i8')
s1.update(d1)

s2 = SpaceSaving(capacity=5, dtype='i8')
s2.update(d2)

empty = SpaceSaving(capacity=5, dtype='i8')

two = SpaceSaving(capacity=7, dtype='i8')
two.update([1, 2])

half = SpaceSaving(capacity=10, dtype='i8')
half.update(d1)

big_1 = SpaceSaving(capacity=20, dtype='i8')
big_1.update(data_i8)

data_i8b = np.random.RandomState(7).gamma(0.1, 0.1, size=10000).round(2) * 100
big_2 = SpaceSaving(capacity=20, dtype='i8')
big_2.update(data_i8b.astype('i8'))

object_1 = SpaceSaving(capacity=20, dtype=object)
object_1.update(data_i8)
object_2 = SpaceSaving(capacity=20, dtype=object)
object_2.update(data_i8b)

summaries = [(s1, s2),
             (empty, empty),
             (s1, s1),
             (s1, empty),
             (empty, s1),
             (empty, two),
             (empty, big_1),
             (two, big_1),
             (big_1, two),
             (half, s1),
             (big_1, big_2),
             (object_1, object_2)]


@pytest.mark.parametrize('s1, s2', summaries)
def test_merge_algorithm(s1, s2):
    s3 = SpaceSaving(capacity=s1.capacity, dtype=s1.dtype)
    s3.merge(s1, s2)
    s4 = copy(s1)
    s4.merge(s2)
    # Equivalent operations
    np.testing.assert_equal(s3.counters(), s4.counters())

    res = [Counter(*i) for i in s3.counters()]
    # Check that the output is sorted
    res2 = sorted(res, key=counter_sort_key, reverse=True)
    assert res == res2
    # Due to equivalent counters maybe sorting differently we just check that
    # the C code returns only counters that *may* have sorted into the topk
    sol = merge(s1, s2)
    assert len(sol) == len(res)
    for c in res:
        assert counter_sort_key(c) >= counter_sort_key(sol[-1])


def test_merge_errors():
    s = SpaceSaving(capacity=10, dtype=object)
    s1 = SpaceSaving(capacity=10, dtype=object)
    s1.update(data_object)

    with pytest.raises(TypeError):
        s.merge(s1, 'not correct type')

    with pytest.raises(ValueError):
        s.merge(s1, big_1)

    # Nothing added before error checking
    assert s.size() == 0



from __future__ import print_function, division, absolute_import

import pickle
from copy import copy

import numpy as np
import pytest

from crick import SummaryStats

normal = np.random.normal(50, scale=100, size=10000)
missing = np.random.randint(0, 10000, size=1000)
normal[missing] = np.nan

empty = np.array([])
one = np.array([1])
duplicate = np.array([1, 1])
different = np.array([1, 2])

datasets = [normal, empty, one, duplicate, different]

RTOL = 0
ATOL = 1e-8


@pytest.mark.parametrize('x', datasets)
def test_basic_stats(x):
    s = SummaryStats()
    s.update(x)

    assert s.count() == np.count_nonzero(~np.isnan(x))
    np.testing.assert_allclose(s.sum(), np.nansum(x), rtol=RTOL, atol=ATOL)
    np.testing.assert_equal(s.min(), np.nanmin(x) if len(x) else np.nan)
    np.testing.assert_equal(s.max(), np.nanmax(x) if len(x) else np.nan)
    np.testing.assert_allclose(s.mean(), np.nanmean(x) if len(x) else np.nan,
                               rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.var(), np.nanvar(x) if len(x) else np.nan,
                               rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.std(), np.nanstd(x) if len(x) else np.nan,
                               rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize('x', datasets)
@pytest.mark.parametrize('bias', [True, False])
def test_skew(x, bias):
    stats = pytest.importorskip('scipy.stats')
    s = SummaryStats()
    s.update(x)
    res = s.skew(bias=bias)
    sol = stats.skew(x[~np.isnan(x)], bias=bias) if len(x) else np.nan
    np.testing.assert_allclose(res, sol, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize('x', datasets)
@pytest.mark.parametrize('bias', [True, False])
@pytest.mark.parametrize('fisher', [True, False])
def test_kurt(x, bias, fisher):
    stats = pytest.importorskip('scipy.stats')
    s = SummaryStats()
    s.update(x)

    res = s.kurt(bias=bias, fisher=fisher)
    if len(x):
        sol = stats.kurtosis(x[~np.isnan(x)], bias=bias, fisher=fisher)
    else:
        sol = np.nan
    np.testing.assert_allclose(res, sol, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize('x', [empty, normal])
def test_pickle(x):
    s = SummaryStats()
    s.update(x)
    s2 = pickle.loads(pickle.dumps(s, protocol=2))
    np.testing.assert_equal(s.count(), s2.count())
    np.testing.assert_equal(s.sum(), s2.sum())
    np.testing.assert_equal(s.min(), s2.min())
    np.testing.assert_equal(s.max(), s2.max())
    np.testing.assert_equal(s.var(), s2.var())
    np.testing.assert_equal(s.skew(), s2.skew())
    np.testing.assert_equal(s.kurt(), s2.kurt())


def test_repr():
    s = SummaryStats()
    assert str(s) == 'SummaryStats<count=0>'
    s.add(10)
    assert str(s) == 'SummaryStats<count=1>'


def test_weights():
    s = SummaryStats()
    s.add(10, 2)
    assert s.count() == 2
    assert s.sum() == 10

    x = np.array([1, 2, 3, 4, 5])
    s.update(x, 2)
    assert s.count() == 12
    assert s.sum() == x.sum() + 10

    s = SummaryStats()
    x = np.array([1, 2, 3, 4, 5])
    s.update(x, x)
    assert s.count() == x.sum()


def test_add_update_errors():
    s = SummaryStats()

    x = np.array([1, 2, 3])
    for c in [-1, 0, np.array([1, 1, -1])]:
        with pytest.raises(ValueError):
            s.update(x, c)

    for c in [-1, 0]:
        with pytest.raises(ValueError):
            s.update(1, c)

    with pytest.raises(ValueError):
        s.update(x, np.nan)

    with pytest.raises(ValueError):
        s.add(1, np.nan)


def test_merge():
    s = SummaryStats()
    half = int(len(normal)/2)
    s.update(normal[:half])
    s2 = SummaryStats()
    s2.update(normal[half:])
    sol = SummaryStats()
    sol.update(normal)

    s.merge(s2)
    np.testing.assert_allclose(s.count(), sol.count(), rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.sum(), sol.sum(), rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.min(), sol.min(), rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.max(), sol.max(), rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.var(), sol.var(), rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.skew(), sol.skew(), rtol=RTOL, atol=ATOL)
    np.testing.assert_allclose(s.kurt(), sol.kurt(), rtol=RTOL, atol=ATOL)

    # Test merge both ways is idempotent
    empty_with_full = SummaryStats()
    empty_with_full.merge(sol)
    full_with_empty = copy(sol)
    full_with_empty.merge(SummaryStats())

    for s in [empty_with_full, full_with_empty]:
        np.testing.assert_equal(s.count(), sol.count())
        np.testing.assert_equal(s.sum(), sol.sum())
        np.testing.assert_equal(s.min(), sol.min())
        np.testing.assert_equal(s.max(), sol.max())
        np.testing.assert_equal(s.var(), sol.var())
        np.testing.assert_equal(s.skew(), sol.skew())
        np.testing.assert_equal(s.kurt(), sol.kurt())



from __future__ import print_function, division, absolute_import

import pickle

import numpy as np
import pytest

from crick import TDigest


# -- Distributions for testing --
N = 100000

# A highly skewed gamma distribution
gamma = np.random.gamma(0.1, 0.1, N)

# Uniform from [0, 1]
uniform = np.random.uniform(0, 1, N)

# A very narrow normal distribution
normal = np.random.normal(0, 1e-5, N)

# Sequential, non-repeated, sorted
sequential = np.arange(N) * 1e-5

# Reversed of above
reverse_sequential = np.arange(N, 0, -1) * 1e-5

# A mix of a narrow normal and a uniform distribution
mixed = np.concatenate([np.random.normal(0, 1e-5, N//2),
                        np.random.uniform(-1, 1, N//2)])
np.random.shuffle(mixed)

# A step function, each value repeated 100 times
step = np.concatenate([np.arange(N/100)] * 100)
np.random.shuffle(step)

# Sorted of above
sorted_step = np.sort(step)


distributions = [gamma, uniform, normal, sequential, reverse_sequential,
                 mixed, step, sorted_step, ]


def quantiles_to_q(data, quant):
    """Convert quantiles of data to the q they represent"""
    N = len(data)
    quant2 = quant[:, None]
    return (2 * (data < quant2).sum(axis=1) +
            (data == quant2).sum(axis=1)) / (2 * N)


def q_to_x(data, q):
    """Convert q to x such that cdf(x) == q"""
    N = len(data)
    sorted_data = np.sort(data)
    x = np.empty_like(q)
    for i in range(len(q)):
        ix = N * q[i] - 0.5
        index = int(ix)
        p = ix - index
        x[i] = sorted_data[index] * (1 - p) + sorted_data[index + 1] * p
    return x


def check_valid_quantile_and_cdf(t):
    min = t.min()
    max = t.max()
    # Quantile
    q = np.linspace(0, 1, 100, endpoint=True)
    res = t.quantile(q)
    assert (np.diff(res) > 0).all()
    assert ((min <= res) & (res <= max)).all()

    x = np.linspace(min, max, 100, endpoint=True)
    res = t.cdf(x)
    assert (np.diff(x) > 0).all()
    assert ((res >= 0) & (res <= 1)).all()


@pytest.mark.parametrize('data', distributions)
def test_distributions(data):
    t = TDigest()
    t.update(data)

    assert t.size() == len(data)
    assert t.min() == data.min()
    assert t.max() == data.max()

    check_valid_quantile_and_cdf(t)

    # *Quantile
    q = np.array([0.001, 0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.99, 0.999])
    est = t.quantile(q)
    q_est = quantiles_to_q(data, est)
    np.testing.assert_allclose(q, q_est, atol=0.012, rtol=0)

    # *CDF
    x = q_to_x(data, q)
    q_est = t.cdf(x)
    np.testing.assert_allclose(q, q_est, atol=0.005)


def test_init():
    t = TDigest(500)
    assert t.compression == 500

    with pytest.raises(TypeError):
        TDigest('foo')

    for c in [np.nan, np.inf, -np.inf]:
        with pytest.raises(ValueError):
            TDigest(c)


def test_repr():
    t = TDigest(500)
    assert str(t) == "TDigest<compression=500.0, size=0.0>"
    t.update(np.arange(100))
    assert str(t) == "TDigest<compression=500.0, size=100.0>"


def test_empty():
    t = TDigest()
    assert t.size() == 0
    assert len(t.centroids()) == 0
    assert np.isnan(t.min())
    assert np.isnan(t.max())
    assert np.isnan(t.quantile(0.5))
    assert np.isnan(t.cdf(0.5))


def test_single():
    t = TDigest()
    t.add(10)
    assert t.min() == 10
    assert t.max() == 10
    assert t.size() == 1

    assert t.quantile(0) == 10
    assert t.quantile(0.5) == 10
    assert t.quantile(1) == 10

    assert t.cdf(9) == 0
    assert t.cdf(10) == 0.5
    assert t.cdf(11) == 1


def test_nonfinite():
    t = TDigest()
    data = gamma.copy()
    data[::10] = np.nan
    data[::7] = np.inf
    t.update(data)
    finite = data[np.isfinite(data)]
    assert t.size() == len(finite)
    assert t.min() == finite.min()
    assert t.max() == finite.max()

    t = TDigest()
    t.add(np.nan)
    t.add(np.inf)
    t.add(-np.inf)
    assert t.size() == 0

    for w in [np.inf, -np.inf, np.nan]:
        t = TDigest()
        with pytest.raises(ValueError):
            t.add(1, w)

        w = np.array([1, 2, w, 3, 4])
        t = TDigest()
        with pytest.raises(ValueError):
            t.update(np.ones(5), w)


def test_small_w():
    eps = np.finfo('f8').eps
    t = TDigest()
    t.update(gamma, eps)
    assert t.size() == 0
    assert len(t.centroids()) == 0

    t = TDigest()
    t.add(1, eps)
    assert t.size() == 0
    assert len(t.centroids()) == 0


def test_update_non_numeric_errors():
    data = np.array(['foo', 'bar', 'baz'])
    t = TDigest()

    with pytest.raises(TypeError):
        t.update(data)

    with pytest.raises(TypeError):
        t.update(1, data)

    with pytest.raises(TypeError):
        t.add('foo')

    with pytest.raises(TypeError):
        t.add(1, 'foo')


def test_quantile_and_cdf_non_numeric():
    t = TDigest()
    t.update(np.arange(5))

    with pytest.raises(TypeError):
        t.quantile('foo')

    with pytest.raises(TypeError):
        t.update(['foo'])

    with pytest.raises(TypeError):
        t.cdf('foo')

    with pytest.raises(TypeError):
        t.cdf(['foo'])


def test_quantile_and_cdf_shape():
    t = TDigest()
    t.update(np.arange(5))

    assert isinstance(t.quantile(0.5), np.float64)
    assert isinstance(t.cdf(2), np.float64)

    res = t.quantile(())
    assert res.shape == (0,)
    res = t.cdf(())
    assert res.shape == (0,)

    qs = [np.array([0.5, 0.9]),
          np.array([[0.5, 0.9], [0, 1]]),
          np.linspace(0, 1, 100)[10:-10:2]]
    for q in qs:
        res = t.quantile(q)
        assert res.shape == q.shape
        res = t.cdf(q)
        assert res.shape == q.shape


def test_histogram():
    t = TDigest()
    data = np.random.normal(size=10000)
    t.update(data)
    hist, bins = t.histogram(100)
    assert len(hist) == 100
    assert len(bins) == 101
    c = t.cdf(bins)
    np.testing.assert_allclose((c[1:] - c[:-1]) * t.size(), hist)

    min = t.min()
    max = t.max()
    eps = np.finfo('f8').eps
    bins = np.array([min - 1, min - eps,
                     min, min + (max - min)/2, max,
                     max + eps, max + 1])
    hist, bins2 = t.histogram(bins)
    np.testing.assert_allclose(bins, bins2)
    assert hist[0] == 0
    assert hist[1] == 0
    assert hist[-2] == 0
    assert hist[-1] == 0
    assert hist.sum() == t.size()

    # range ignored when bins provided
    hist2, bins2 = t.histogram(bins, range=(-5, -3))
    np.testing.assert_allclose(hist, hist2)
    np.testing.assert_allclose(bins, bins2)


def test_histogram_small_n():
    t = TDigest()
    t.add(1)

    hist, bins = t.histogram(10)
    assert len(hist) == 10
    assert len(bins) == 11
    assert bins[0] == 0.5
    assert bins[-1] == 1.5
    assert hist.sum() == 1

    t.add(2)
    hist, bins = t.histogram(10)
    assert hist.sum() == 2
    assert bins[0] == 1
    assert bins[-1] == 2

    hist, bins = t.histogram(range=(-5, -3))
    assert hist.sum() == 0


def test_histogram_empty():
    t = TDigest()
    for b, r in [(5, None), (5, (-1, 1)), (np.arange(6), None)]:
        hist, bins = t.histogram(bins=b, range=r)
        assert len(hist) == 5
        assert len(bins) == 6
        if r is not None:
            assert bins[0] == r[0]
            assert bins[-1] == r[1]
        assert (hist == 0).all()
        assert (np.diff(bins) > 0).all()


def test_histogram_errors():
    t = TDigest()
    t.update(np.random.uniform(1000))

    for r in [('a', 'b'), 1]:
        with pytest.raises(TypeError):
            t.histogram(range=r)

    with pytest.raises(Exception):
        t.histogram(range=1)

    for r in [(np.nan, 1), (np.inf, 1), (1, np.nan), (1, np.inf)]:
        with pytest.raises(ValueError):
            t.histogram(range=r)

    with pytest.raises(ValueError):
        t.histogram(range=(1, 0))

    for b in ['a', -1, np.arange(4).reshape((2, 2)),
              np.arange(0, 10, -1), np.array([np.nan, 0, 1])]:
        with pytest.raises(ValueError):
            t.histogram(bins=b)


def test_weights():
    t = TDigest()
    t.add(1, 10)
    assert t.size() == 10

    x = np.arange(5)
    w = np.array([1, 2, 1, 2, 1])

    t = TDigest()
    t.update(x, 10)
    assert t.size() == len(x) * 10

    t = TDigest()
    t.update(x, w)
    assert t.size() == w.sum()


def test_serialize():
    not_empty = TDigest()
    not_empty.update(gamma)
    empty = TDigest()
    for t in [not_empty, empty]:
        t2 = pickle.loads(pickle.dumps(t))
        assert t.compression == t2.compression
        assert (t.centroids() == t2.centroids()).all()
        np.testing.assert_equal(t.min(), t2.min())
        np.testing.assert_equal(t.max(), t2.max())
        np.testing.assert_equal(t.size(), t2.size())


def test_merge():
    t = TDigest()
    t2 = TDigest()
    t3 = TDigest()
    a = np.random.uniform(0, 1, N)
    b = np.random.uniform(2, 3, N)
    data = np.concatenate([a, b])
    t2.update(a)
    t3.update(b)

    t2_centroids = t2.centroids()

    t.merge(t2, t3)
    assert t.min() == min(t2.min(), t3.min())
    assert t.max() == max(t2.max(), t3.max())
    assert t.size() == t2.size() + t3.size()
    # Check no mutation of args
    assert (t2.centroids() == t2_centroids).all()

    # *Quantile
    q = np.array([0.001, 0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.99, 0.999])
    est = t.quantile(q)
    q_est = quantiles_to_q(data, est)
    np.testing.assert_allclose(q, q_est, atol=0.012, rtol=0)

    # *CDF
    x = q_to_x(data, q)
    q_est = t.cdf(x)
    np.testing.assert_allclose(q, q_est, atol=0.005)

    with pytest.raises(TypeError):
        t.merge(t2, 'not a tdigest')


def test_scale():
    t = TDigest()
    t.update(uniform)

    for factor in [0.5, 2]:
        t2 = t.scale(factor)
        assert t is not t2
        assert t.size() * factor == t2.size()
        assert t.min() == t2.min()
        assert t.max() == t2.max()
        a = t.centroids()
        b = t2.centroids()
        np.testing.assert_array_equal(a['mean'], b['mean'])
        np.testing.assert_allclose(a['weight'] * factor, b['weight'])

    for val in [-0.5, 0, np.nan, np.inf]:
        with pytest.raises(ValueError):
            t.scale(val)

    with pytest.raises(TypeError):
        t.scale('foobar')

    # Test scale compacts
    eps = np.finfo('f8').eps
    t = TDigest()
    t.update([1, 2, 3, 4, 5],
             [1, 1000, 1, 10000, 1])
    t2 = t.scale(eps)
    assert len(t2.centroids()) == 2

    # Compacts to 0
    t = TDigest()
    t.update([1, 2, 3, 4, 5])
    t2 = t.scale(eps)
    assert len(t2.centroids()) == 0



from __future__ import absolute_import

from .tdigest import TDigest
from .space_saving import SpaceSaving
from .stats import SummaryStats

from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions
